#include "Temporizador.h"

//** CONSTRUCTOR **
  Temporizador::Temporizador(uint8_t me_mo, uint8_t ti_in_es, uint8_t ti_ar, uint8_t ti_ar_es, uint8_t ti_ac, uint8_t ti_ac_es, uint8_t st_te, uint8_t pe_ti, uint8_t re_pi)
  {
    _init();
    _setup(me_mo, ti_in_es, ti_ar, ti_ar_es, ti_ac, ti_ac_es, st_te, pe_ti, re_pi);
  }

  Temporizador::Temporizador(uint8_t me_mo, uint8_t i_aut, uint8_t r_enc, uint8_t ti_in_es, uint8_t ti_ar, uint8_t ti_ar_es, uint8_t ti_ac, uint8_t ti_ac_es, uint8_t st_te, uint8_t pe_ti, uint8_t re_pi)
  {
    _init();
    setEepromAddressInicioAutomatico(i_aut);
    setEepromAdressReanudarEncendido(r_enc);
    _setup(me_mo, ti_in_es, ti_ar, ti_ar_es, ti_ac, ti_ac_es, st_te, pe_ti, re_pi);
  }


//** DESTRUCTOR **
  Temporizador::~Temporizador()
  {

  }


//** MÉTODOS PRIVADOS**
  // Para constructor
    void Temporizador::_init()
    {
      _instancia_num = ++_instancias_num;

      if(!_fm24c64_i2c_address)
        _fm24c64_i2c_address = FMRAM_DEFAULT_ADDRESS;

      if(!_ds3231_i2c_address)
        _ds3231_i2c_address = RTC_DEFAULT_ADDRESS;
    }

    void Temporizador::_setup(uint8_t me_mo, uint8_t ti_in_es, uint8_t ti_ar, uint8_t ti_ar_es, uint8_t ti_ac, uint8_t ti_ac_es, uint8_t st_te, uint8_t pe_ti, uint8_t re_pi)
    {
      setMemoryMode(me_mo);

      setEepromAddressTiempoInicioEspera(ti_in_es);
      setEepromAddressTiempoArranque(ti_ar);
      setEepromAddressTiempoArranqueEspera(ti_ar_es);
      setEepromAddressTiempoEncendido(ti_ac);
      setEepromAddressTiempoEncendidoEspera(ti_ac_es);

      setMemoryAddressPeriodo(st_te);
      setMemoryAddressPeriodoTime(pe_ti);

      setRelayPin(re_pi);
    }


  // EEPROM
    void Temporizador::EEPROMWritelong(uint8_t address, uint32_t value)
    {
      //Decomposition from a long to 4 bytes by using bitshift.
      //One = Most significant -> Four = Least significant byte
      uint8_t four  = (value & 0xFF);
      uint8_t three = ((value >> 8) & 0xFF);
      uint8_t two   = ((value >> 16) & 0xFF);
      uint8_t one   = ((value >> 24) & 0xFF);

      //Write the 4 bytes into the eeprom memory.
      EEPROM.write(address, four);
      EEPROM.write(address + 1, three);
      EEPROM.write(address + 2, two);
      EEPROM.write(address + 3, one);
    }

    uint32_t Temporizador::EEPROMReadlong(uint8_t address)
    {
      //Read the 4 bytes from the eeprom memory.
      uint32_t four  = EEPROM.read(address);
      uint32_t three = EEPROM.read(address + 1);
      uint32_t two   = EEPROM.read(address + 2);
      uint32_t one   = EEPROM.read(address + 3);
      
      //Return the recomposed long by using bitshift.
      return ((four << 0) & 0xFF) + ((three << 8) & 0xFFFF) + ((two << 16) & 0xFFFFFF) + ((one << 24) & 0xFFFFFFFF);
    }


  // VRAM / FMRAM
    void Temporizador::writeMemory(uint8_t addr, uint8_t values)
    {
      writeMemory(addr, values, 1);
    }

    void Temporizador::writeMemory(uint8_t addr, uint32_t values, uint8_t nBytes)
    {
      uint8_t i2c_address = _memory_mode == NVRAM_MODE? _ds3231_i2c_address: _fm24c64_i2c_address;

      if(_memory_mode == NVRAM_MODE)
        addr = 0x07 + addr;

      Wire.beginTransmission(i2c_address);
      Wire.write(addr); //set storage register to begin
      Wire.write((uint8_t*)&values, nBytes);//write 4 bytes
      Wire.endTransmission();
    }


    uint8_t Temporizador::readMemory(uint8_t addr)
    {
      return readMemory(addr, 1);
    }

    uint32_t Temporizador::readMemory(uint8_t addr, uint16_t nBytes)
    {
      uint8_t i2c_address = _memory_mode == NVRAM_MODE? _ds3231_i2c_address: _fm24c64_i2c_address;

      if(_memory_mode == NVRAM_MODE)
        addr = 0x07 + addr;

      Wire.beginTransmission(i2c_address);
      Wire.write(addr); //set storage register to begin
      Wire.endTransmission();

      uint32_t value;

      Wire.requestFrom(i2c_address, nBytes);
      Wire.readBytes((uint8_t*)&value, nBytes); //read 4 bytes
      Wire.endTransmission();

      return value;
    }


  // Proximo periodo
    void Temporizador::proximoPeriodo()
    {
      switch (getPeriodo()){
        case STATUS_INICIANDO:
          _periodo      = STATUS_INICIO_ESPERA;
          _periodo_time = getTiempoInicioEspera();
          break;

        case STATUS_INICIO_ESPERA:
          _periodo      = STATUS_ARRANQUE;
          _periodo_time = getTiempoArranque();
          break;

        case STATUS_ARRANQUE:
          _periodo      = STATUS_ARRANQUE_ESPERA;
          _periodo_time = getTiempoArranqueEspera();
          break;

        case STATUS_ARRANQUE_ESPERA:
          _periodo      = STATUS_ENCENDIDO;
          _periodo_time = getTiempoEncendido();
          break;

        case STATUS_ENCENDIDO:
          _periodo      = STATUS_ENCENDIDO_ESPERA;
          _periodo_time = getTiempoEncendidoEspera();
          break;

        case STATUS_ENCENDIDO_ESPERA:
          _periodo      = STATUS_ENCENDIDO;
          _periodo_time = getTiempoEncendido();
          break;
      }


      // Guardamos en memoria
      writeMemory(_memory_address_periodo, _periodo);
      writeMemory(_memory_address_periodo_time, _periodo_time, 4);


      /*
      // Si Retraso o Arranque igual a 0, recalculamos
      if((status_temporizador == STATUS_RETARDO || status_temporizador == STATUS_ARRANQUE) && periodo_time == 0)
        procesar_proximo_periodo();
      */
    }


//** MÉTODOS PUBLICOS **
  // Set Fecha y Hora
    void Temporizador::setFecha(uint16_t anho, uint8_t mes, uint8_t dia)
    {
      _RTC.adjust(DateTime(anho, mes, dia, _now.hour(), _now.minute(), _now.second()));
    }

    void Temporizador::setHora(uint8_t hrs, uint8_t min, uint8_t seg)
    {
      _RTC.adjust(DateTime(_now.year(), _now.month(), _now.day(), hrs, min, seg));
    }

    void Temporizador::setFechaHora(uint16_t anho, uint8_t mes, uint8_t dia, uint8_t hrs, uint8_t min, uint8_t seg)
    {
      _RTC.adjust(DateTime(anho, mes, dia, hrs, min, seg));
    }


  // Get Fecha y Hora
    String Temporizador::getFecha()
    {
      String year  = (String) _now.year();
      String month = (_now.month() < 10? "0": "") + (String) _now.month();
      String day   = (_now.day() < 10? "0": "") + (String) _now.day();

      return day + "/" + month + '/' + year;
    }

    String Temporizador::getHora()
    {
      String hour   = (_now.hour() < 10? "0": "") + (String) _now.hour();
      String minute = (_now.minute() < 10? "0": "") + (String) _now.minute();
      String second = (_now.second() < 10? "0": "") + (String) _now.second();

      return hour + ":" + minute + ':' + second;
    }


  // Configuraciones Hardware
    void Temporizador::setRelayPin(uint8_t relay_pin)
    {
      _relay_pin = relay_pin;
    }

    void Temporizador::setMemoryMode(uint8_t memory_mode)
    {
      _memory_mode = memory_mode;
    }

    void Temporizador::setRtcAdress(uint8_t address)
    {
      _ds3231_i2c_address = address;
    }

    void Temporizador::setFmramAdress(uint8_t address)
    {
      _fm24c64_i2c_address = address;
    }

    boolean Temporizador::getStatusRelay()
    {
      return _status_relay;
    }

    boolean Temporizador::getStatusDetenido()
    {
      return _detenido;
    }


  // Set EEPROM address
    void Temporizador::setEepromAddressInicioAutomatico(uint8_t address)
    {
      _eeprom_address_inicio_automatico = address;
    }

    void Temporizador::setEepromAdressReanudarEncendido(uint8_t address)
    {
      _eeprom_address_reanudar_encendido = address;
    }

    void Temporizador::setEepromAddressTiempoInicioEspera(uint8_t address)
    {
      _eeprom_address_tiempo_inicio_espera = address;
    }

    void Temporizador::setEepromAddressTiempoArranque(uint8_t address)
    {
      _eeprom_address_tiempo_arranque = address;
    }

    void Temporizador::setEepromAddressTiempoArranqueEspera(uint8_t address)
    {
      _eeprom_address_tiempo_arranque_espera = address;
    }

    void Temporizador::setEepromAddressTiempoEncendido(uint8_t address)
    {
      _eeprom_address_tiempo_activo = address;
    }

    void Temporizador::setEepromAddressTiempoEncendidoEspera(uint8_t address)
    {
      _eeprom_address_tiempo_activo_espera = address;
    }


  // Set Memory Address
    void Temporizador::setMemoryAddressPeriodo(uint8_t address)
    {
      _memory_address_periodo = address;
    }

    void Temporizador::setMemoryAddressPeriodoTime(uint8_t address)
    {
      _memory_address_periodo_time = address;
    }


  // Set Configuraciones Temporizador
    void Temporizador::setInicioAutomatico(uint8_t inicio_automatico)
    {
      EEPROM.write(_eeprom_address_inicio_automatico, inicio_automatico);
    }

    void Temporizador::setReanudarEncendido(uint8_t reanudar_encendido)
    {
      EEPROM.write(_eeprom_address_reanudar_encendido, reanudar_encendido);
    }

    void Temporizador::setTiempoInicioEspera(uint32_t segundos)
    {
      EEPROMWritelong(_eeprom_address_tiempo_inicio_espera, segundos);
    }

    void Temporizador::setTiempoInicioEspera(uint8_t minutos, uint8_t segundos)
    {
      uint32_t _segundos = (minutos * 60) + segundos;

      EEPROMWritelong(_eeprom_address_tiempo_inicio_espera, _segundos);
    }

    void Temporizador::setTiempoArranque(uint32_t segundos)
    {
      EEPROMWritelong(_eeprom_address_tiempo_arranque, segundos);
    }

    void Temporizador::setTiempoArranque(uint8_t horas, uint8_t minutos, uint8_t segundos)
    {
      uint32_t _segundos = (horas * 3600) + (minutos * 60) + segundos;

      EEPROMWritelong(_eeprom_address_tiempo_arranque, _segundos);
    }

    void Temporizador::setTiempoArranqueEspera(uint32_t segundos)
    {
      EEPROMWritelong(_eeprom_address_tiempo_arranque_espera, segundos);
    }

    void Temporizador::setTiempoArranqueEspera(uint8_t horas, uint8_t minutos, uint8_t segundos)
    {
      uint32_t _segundos = (horas * 3600) + (minutos * 60) + segundos;

      EEPROMWritelong(_eeprom_address_tiempo_arranque_espera, _segundos);
    }

    void Temporizador::setTiempoEncendido(uint32_t segundos)
    {
      EEPROMWritelong(_eeprom_address_tiempo_activo, segundos);
    }

    void Temporizador::setTiempoEncendido(uint8_t horas, uint8_t minutos, uint8_t segundos)
    {
      uint32_t _segundos = (horas * 3600) + (minutos * 60) + segundos;

      EEPROMWritelong(_eeprom_address_tiempo_activo, _segundos);
    }

    void Temporizador::setTiempoEncendidoEspera(uint32_t segundos)
    {
      EEPROMWritelong(_eeprom_address_tiempo_activo_espera, segundos);
    }

    void Temporizador::setTiempoEncendidoEspera(uint8_t horas, uint8_t minutos, uint8_t segundos)
    {
      uint32_t _segundos = (horas * 3600) + (minutos * 60) + segundos;

      EEPROMWritelong(_eeprom_address_tiempo_activo_espera, _segundos);
    }


  // Get Temporizador EEPROM
    boolean Temporizador::getInicioAutomatico()
    {
      return boolean(EEPROM.read(_eeprom_address_inicio_automatico));
    }

    boolean Temporizador::getReanudarInicio()
    {
      return boolean(EEPROM.read(_eeprom_address_reanudar_encendido));
    }

    uint32_t Temporizador::getTiempoInicioEspera()
    {
      return EEPROMReadlong(_eeprom_address_tiempo_inicio_espera);
    }

    uint32_t Temporizador::getTiempoArranque()
    {
      return EEPROMReadlong(_eeprom_address_tiempo_arranque);
    }

    uint32_t Temporizador::getTiempoArranqueEspera()
    {
      return EEPROMReadlong(_eeprom_address_tiempo_arranque_espera);
    }

    uint32_t Temporizador::getTiempoEncendido()
    {
      return EEPROMReadlong(_eeprom_address_tiempo_activo);
    }

    uint32_t Temporizador::getTiempoEncendidoEspera()
    {
      return EEPROMReadlong(_eeprom_address_tiempo_activo_espera);
    }


  // Get Status Temporizador
    uint8_t Temporizador::getInstanciasNum()
    {
      return _instancias_num;
    }

    uint8_t Temporizador::getInstanciaNum()
    {
      return _instancia_num;
    }

    uint8_t Temporizador::getPeriodo()
    {
      return _periodo;
    }

    uint32_t Temporizador::getPeriodoTiempo()
    {
      return _periodo_time;
    }


  // Reinicio y detener del temporizador
    void Temporizador::setDefaults()
    {
      setInicioAutomatico(1);
      setTiempoInicioEspera(10);
      setTiempoArranque(10);
      setTiempoArranqueEspera(10);
      setTiempoEncendido(10);
      setTiempoEncendidoEspera(10);
    }

    void Temporizador::reiniciar()
    {
      _periodo      = STATUS_INICIANDO;
      _periodo_time = 0;

      // Guardamos en memoria
      writeMemory(_memory_address_periodo, _periodo);
      writeMemory(_memory_address_periodo_time, _periodo_time, 4);

      proximoPeriodo();
    }

    void Temporizador::deshabilitar()
    {
      _periodo      = STATUS_DESHABILITADO;
      _periodo_time = 0;

      // Guardamos en memoria
      writeMemory(_memory_address_periodo, _periodo);
      writeMemory(_memory_address_periodo_time, _periodo_time, 4);
    }

    void Temporizador::detener()
    {
      _detenido = true;
    }

    void Temporizador::reanudar()
    {
      _detenido = false;
    }


  // PROCESO DEL TEMPORIZADOR
    void Temporizador::inizializar()
    {
      if(_instancia_num == 1){
        Wire.begin();

        if(!_RTC.begin()) {
            if(Serial)
              Serial.println("Couldn't find RTC");
        }

        if(_RTC.lostPower()){
          if(Serial)
            Serial.println("RTC lost power, lets set the time!");

          //_RTC.adjust(DateTime(__DATE__, __TIME__));
          _RTC.adjust(DateTime(2018, 1, 1, 0, 0, 0));
        }

        _now = _RTC.now();
      }

      if(getReanudarInicio()){
        _periodo      = readMemory(_memory_address_periodo);
        _periodo_time = readMemory(_memory_address_periodo_time, 4);
        _periodo_time++;

      }else{
        _periodo      = STATUS_INICIANDO;
        _periodo_time = 0;
      }

      _detenido = getInicioAutomatico() == false;

      pinMode(_relay_pin, OUTPUT);
    }

    void Temporizador::procesarTemporizador()
    {
      if(_instancia_num == 1)
        _now = _RTC.now();

      // Verificamos RTC Transcura bien
      if(millis() / 1000 >= _T[0] +1){
        if(_now.unixtime() == _T[1]){
          _periodo      = STATUS_RTC_ERROR;
          _periodo_time = 0;
        }

        _T[0] = millis() / 1000;
        _T[1] = _now.unixtime();
      }

      // Monitoreo del Temporizador
      if(!getStatusDetenido() && _periodo >= STATUS_INICIANDO){
        // Una vez cada segundo por Instancia
        if(_now.unixtime() >= _T[2] +1) {
          writeMemory(_memory_address_periodo_time, --_periodo_time, 4);

          if(_periodo_time <= 0)
            proximoPeriodo();

          // Estatus del relevador
          _status_relay = (_periodo == STATUS_ARRANQUE || _periodo == STATUS_ENCENDIDO) && _periodo_time > 0;
          digitalWrite(_relay_pin, _status_relay);

          _T[2] = _now.unixtime();
        }

      }else if(_status_relay){
        _status_relay = false;
        digitalWrite(_relay_pin, _status_relay);
      }
    }
